#ifndef __SYNC_H__
#define __SYNC_H__


#ifdef _WIN32
#include <windows.h>
#include <assert.h>

#define THREAD_PROC WINAPI

class haMutex { 
    CRITICAL_SECTION cs;
  public:
    haMutex() { 
	InitializeCriticalSection(&cs);
    }
    ~haMutex() { 
	DeleteCriticalSection(&cs);
    }
    void lock() { 
	EnterCriticalSection(&cs);
    }
    void unlock() { 
	LeaveCriticalSection(&cs);
    }
};


class haThread { 
    HANDLE h;
  public:
    void run(void (THREAD_PROC* func)(void*), void* arg) { 
	DWORD threadid;
	h = CreateThread(NULL, NULL, LPTHREAD_START_ROUTINE(func), arg,
			 0, &threadid);
    }

    void join() { 
	WaitForSingleObject(h, INFINITE);
	CloseHandle(h);
	h = NULL;
    }
};

class haEvent { 
    HANDLE e;
  public:
    void wait(haMutex& mutex) { 
	mutex.unlock();
	int rc = WaitForSingleObject(e, INFINITE);
	assert(rc == WAIT_OBJECT_0);
	mutex.lock();
    }

    void signal() {
	SetEvent(e);
    }
    
    haEvent() { 
	e = CreateEvent(NULL, false, false, NULL);
    }
    ~haEvent() {
	CloseHandle(e);
    }
};

#else

#include <stddef.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <assert.h>

#define THREAD_PROC 

class haMutex { 
    friend class haEvent;
    pthread_mutex_t cs;
  public:
    haMutex() { 
	pthread_mutex_init(&cs, NULL);
    }
    ~haMutex() { 
	pthread_mutex_destroy(&cs);
    }
    void lock() { 
	pthread_mutex_lock(&cs);
    }
    void unlock() { 
	pthread_mutex_unlock(&cs);
    }
};


class haThread { 
    pthread_t thread;
  public:
    void run(void (THREAD_PROC* func)(void*), void* arg) { 
	pthread_create(&thread, NULL, (void*(*)(void*))func, arg);
    }

    void join() { 
	void* result;
	pthread_join(thread, &result);
    }
};

class haEvent { 
    pthread_cond_t cond;
    int            signaled;
  public:
    void wait(haMutex& mutex) { 
	while (!signaled) {
	    pthread_cond_wait(&cond, &mutex.cs);
	}
	signaled = false;
    }

    void signal() {
	signaled = true;
	pthread_cond_signal(&cond);
    }
    
    haEvent() { 
	pthread_cond_init(&cond, NULL);
	signaled = false;
    }
    ~haEvent() {
	pthread_cond_destroy(&cond);
    }
};

#endif

#endif










